# coding: utf-8
# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Embedding composites for various types of problems and application.
For example:

* :class:`.EmbeddingComposite` for a problem with arbitrary structure that likely
  requires hueristic embedding.
* :class:`.AutoEmbeddingComposite` can save unnecessary embedding for
  problems that might have a structure similar to the child sampler.
* :class:`.LazyFixedEmbeddingComposite` can benefit applications that
  resubmit a BQM with changes in some values.
"""

import itertools

from warnings import warn

import dimod
import minorminer
import functools

from dwave.embedding import (target_to_source, unembed_sampleset, embed_bqm,
                             chain_to_quadratic, EmbeddedStructure)
from dwave.system.warnings import WarningHandler, WarningAction

__all__ = ('EmbeddingComposite',
           'FixedEmbeddingComposite',
           'LazyFixedEmbeddingComposite',
           'LazyEmbeddingComposite',  # deprecated
           'AutoEmbeddingComposite',
           )


class EmbeddingComposite(dimod.ComposedSampler):
    """Maps problems to a structured sampler.

    Automatically minor-embeds a problem into a structured sampler such as a
    D-Wave system. A new minor-embedding is calculated each time one of its
    sampling methods is called.

    Args:
        child_sampler (:class:`dimod.Sampler`):
            A dimod sampler, such as a :obj:`.DWaveSampler`, that accepts
            only binary quadratic models of a particular structure.

        find_embedding (function, optional):
            A function `find_embedding(S, T, **kwargs)` where `S` and `T`
            are edgelists. The function can accept additional keyword arguments.
            Defaults to :func:`minorminer.find_embedding`.

        embedding_parameters (dict, optional):
            If provided, parameters are passed to the embedding method as
            keyword arguments.

        scale_aware (bool, optional, default=False):
            Pass chain interactions to child samplers that accept an `ignored_interactions`
            parameter.

        child_structure_search (function, optional):
            A function `child_structure_search(sampler)` that accepts a sampler
            and returns the :attr:`dimod.Structured.structure`.
            Defaults to :func:`dimod.child_structure_dfs`.

    Examples:

       >>> from dwave.system import DWaveSampler, EmbeddingComposite
       ...
       >>> sampler = EmbeddingComposite(DWaveSampler())
       >>> h = {'a': -1., 'b': 2}
       >>> J = {('a', 'b'): 1.5}
       >>> sampleset = sampler.sample_ising(h, J, num_reads=100)
       >>> sampleset.first.energy
       -4.5


    """
    def __init__(self, child_sampler,
                 find_embedding=minorminer.find_embedding,
                 embedding_parameters=None,
                 scale_aware=False,
                 child_structure_search=dimod.child_structure_dfs):

        self.children = [child_sampler]

        # keep any embedding parameters around until later, because we might
        # want to overwrite them
        self.embedding_parameters = embedding_parameters or {}
        self.find_embedding = find_embedding

        # set the parameters
        self.parameters = parameters = child_sampler.parameters.copy()
        parameters.update(chain_strength=[],
                          chain_break_method=[],
                          chain_break_fraction=[],
                          embedding_parameters=[],
                          return_embedding=[],
                          warnings=[],
                          )

        # set the properties
        self.properties = dict(child_properties=child_sampler.properties.copy())

        # track the child's structure. We use a dfs in case intermediate
        # composites are not structured. We could expose multiple different
        # searches but since (as of 14 june 2019) all composites have single
        # children, just doing dfs seems safe for now.
        self.target_structure = child_structure_search(child_sampler)

        self.scale_aware = bool(scale_aware)

    parameters = None  # overwritten by init
    """dict[str, list]: Parameters in the form of a dict.

    For an instantiated composed sampler, keys are the keyword parameters
    accepted by the child sampler and parameters added by the composite.
    """

    children = None  # overwritten by init
    """list [child_sampler]: List containing the structured sampler."""

    properties = None  # overwritten by init
    """dict: Properties in the form of a dict.

    Contains the properties of the child sampler.
    """

    return_embedding_default = False
    """Defines the default behaviour for :meth:`.sample`'s `return_embedding`
    kwarg.
    """

    warnings_default = WarningAction.IGNORE
    """Defines the default behavior for :meth:`.sample`'s `warnings` kwarg.
    """

    def sample(self, bqm, chain_strength=None,
               chain_break_method=None,
               chain_break_fraction=True,
               embedding_parameters=None,
               return_embedding=None,
               warnings=None,
               **parameters):
        """Sample from the provided binary quadratic model.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            chain_strength (float/mapping/callable, optional):
                Sets the coupling strength between qubits representing variables 
                that form a :term:`chain`. Mappings should specify the required 
                chain strength for each variable. Callables should accept the BQM 
                and embedding and return a float or mapping. By default, 
                `chain_strength` is calculated with
                :func:`~dwave.embedding.chain_strength.uniform_torque_compensation`.

            chain_break_method (function/list, optional):
                Method or methods used to resolve chain breaks. If multiple
                methods are given, the results are concatenated and a new field
                called "chain_break_method" specifying the index of the method
                is appended to the sample set.
                See :func:`~dwave.embedding.unembed_sampleset` and
                :mod:`dwave.embedding.chain_breaks`.

            chain_break_fraction (bool, optional, default=True):
                Add a `chain_break_fraction` field to the unembedded response with
                the fraction of chains broken before unembedding.

            embedding_parameters (dict, optional):
                If provided, parameters are passed to the embedding method as
                keyword arguments. Overrides any `embedding_parameters` passed
                to the constructor.

            return_embedding (bool, optional):
                If True, the embedding, chain strength, chain break method and
                embedding parameters are added to :attr:`dimod.SampleSet.info`
                of the returned sample set. The default behaviour is defined
                by :attr:`return_embedding_default`, which itself defaults to
                False.

            warnings (:class:`~dwave.system.warnings.WarningAction`, optional):
                Defines what warning action to take, if any. See
                :mod:`~dwave.system.warnings`. The default behaviour is defined
                by :attr:`warnings_default`, which itself defaults to
                :class:`~dwave.system.warnings.IGNORE`

            **parameters:
                Parameters for the sampling method, specified by the child
                sampler.

        Returns:
            :obj:`dimod.SampleSet`

        Examples:
            See the example in :class:`EmbeddingComposite`.

        """
        if return_embedding is None:
            return_embedding = self.return_embedding_default

        # solve the problem on the child system
        child = self.child

        # apply the embedding to the given problem to map it to the child sampler
        __, target_edgelist, target_adjacency = self.target_structure

        # add self-loops to edgelist to handle singleton variables
        source_edgelist = list(bqm.quadratic) + [(v, v) for v in bqm.linear]

        # get the embedding
        if embedding_parameters is None:
            embedding_parameters = self.embedding_parameters
        else:
            # we want the parameters provided to the constructor, updated with
            # the ones provided to the sample method. To avoid the extra copy
            # we do an update, avoiding the keys that would overwrite the
            # sample-level embedding parameters
            embedding_parameters.update((key, val)
                                        for key, val in self.embedding_parameters
                                        if key not in embedding_parameters)

        embedding = self.find_embedding(source_edgelist, target_edgelist,
                                        **embedding_parameters)

        if bqm and not embedding:
            raise ValueError("no embedding found")

        if not hasattr(embedding, 'embed_bqm'):
            embedding = EmbeddedStructure(target_edgelist, embedding)

        bqm_embedded = embedding.embed_bqm(bqm, chain_strength=chain_strength,
                                           smear_vartype=dimod.SPIN)

        if warnings is None:
            warnings = self.warnings_default
        elif 'warnings' in child.parameters:
            parameters.update(warnings=warnings)

        warninghandler = WarningHandler(warnings)

        warninghandler.chain_strength(bqm, embedding.chain_strength, embedding)
        warninghandler.chain_length(embedding)

        if 'initial_state' in parameters:
            # if initial_state was provided in terms of the source BQM, we want
            # to modify it to now provide the initial state for the target BQM.
            # we do this by spreading the initial state values over the
            # chains
            state = parameters['initial_state']
            parameters['initial_state'] = {u: state[v]
                                           for v, chain in embedding.items()
                                           for u in chain}

        if self.scale_aware and 'ignored_interactions' in child.parameters:

            ignored = []
            for chain in embedding.values():
                # just use 0 as a null value because we don't actually need
                # the biases, just the interactions
                ignored.extend(chain_to_quadratic(chain, target_adjacency, 0))

            parameters['ignored_interactions'] = ignored

        response = child.sample(bqm_embedded, **parameters)

        def async_unembed(response):
            # unembed the sampleset aysnchronously.

            warninghandler.chain_break(response, embedding)

            sampleset = unembed_sampleset(response, embedding, source_bqm=bqm,
                                          chain_break_method=chain_break_method,
                                          chain_break_fraction=chain_break_fraction,
                                          return_embedding=return_embedding)

            if return_embedding:
                sampleset.info['embedding_context'].update(
                    embedding_parameters=embedding_parameters,
                    chain_strength=embedding.chain_strength)

            if chain_break_fraction and len(sampleset):
                warninghandler.issue("All samples have broken chains",
                                     func=lambda: (sampleset.record.chain_break_fraction.all(), None))

            if warninghandler.action is WarningAction.SAVE:
                # we're done with the warning handler so we can just pass the list
                # off, if later we want to pass in a handler or similar we should
                # do a copy
                sampleset.info.setdefault('warnings', []).extend(warninghandler.saved)

            return sampleset

        return dimod.SampleSet.from_future(response, async_unembed)


class LazyFixedEmbeddingComposite(EmbeddingComposite, dimod.Structured):
    """Maps problems to the structure of its first given problem.

    This composite reuses the minor-embedding found for its first given problem
    without recalculating a new minor-embedding for subsequent calls of its
    sampling methods.

    Args:
        child_sampler (dimod.Sampler):
            Structured dimod sampler.

        find_embedding (function, default=:func:`minorminer.find_embedding`):
            A function `find_embedding(S, T, **kwargs)` where `S` and `T`
            are edgelists. The function can accept additional keyword arguments.
            The function is used to find the embedding for the first problem
            solved.

        embedding_parameters (dict, optional):
            If provided, parameters are passed to the embedding method as keyword
            arguments.

    Examples:

        >>> from dwave.system import LazyFixedEmbeddingComposite, DWaveSampler
        ...
        >>> sampler = LazyFixedEmbeddingComposite(DWaveSampler())
        >>> sampler.nodelist is None  # no structure prior to first sampling
        True
        >>> __ = sampler.sample_ising({}, {('a', 'b'): -1})
        >>> sampler.nodelist  # has structure based on given problem
        ['a', 'b']

    """

    @property
    def nodelist(self):
        """list: Nodes available to the composed sampler."""
        try:
            return self._nodelist
        except AttributeError:
            pass

        if self.adjacency is None:
            return None

        self._nodelist = nodelist = list(self.adjacency)

        # makes it a lot easier for the user if the list can be sorted, so we
        # try
        try:
            nodelist.sort()
        except TypeError:
            # python3 cannot sort unlike types
            pass

        return nodelist

    @property
    def edgelist(self):
        """list: Edges available to the composed sampler."""
        try:
            return self._edgelist
        except AttributeError:
            pass

        adj = self.adjacency

        if adj is None:
            return None

        # remove duplicates by putting into a set
        edges = set()
        for u in adj:
            for v in adj[u]:
                try:
                    edge = (u, v) if u <= v else (v, u)
                except TypeError:
                    # Py3 does not allow sorting of unlike types
                    if (v, u) in edges:
                        continue
                    edge = (u, v)

                edges.add(edge)

        self._edgelist = edgelist = list(edges)

        # makes it a lot easier for the user if the list can be sorted, so we
        # try
        try:
            edgelist.sort()
        except TypeError:
            # python3 cannot sort unlike types
            pass

        return edgelist

    @property
    def adjacency(self):
        """dict[variable, set]: Adjacency structure for the composed sampler."""
        try:
            return self._adjacency
        except AttributeError:
            pass

        if self.embedding is None:
            return None

        self._adjacency = adj = target_to_source(self.target_structure.adjacency,
                                                 self.embedding)

        return adj

    embedding = None
    """Embedding used to map binary quadratic models to the child sampler."""

    def _fix_embedding(self, embedding):
        target_edgelist = self.target_structure.edgelist
        embedding = EmbeddedStructure(target_edgelist, embedding)

        # save the embedding and overwrite the find_embedding function
        self.embedding = embedding
        self.properties.update(embedding=embedding)

        def find_embedding(S, T):
            return embedding

        self.find_embedding = find_embedding

    def sample(self, bqm, **parameters):
        """Sample the binary quadratic model.

        On the first call of a sampling method, finds a :term:`minor-embedding`
        for the given binary quadratic model (BQM). All subsequent calls to its
        sampling methods reuse this embedding.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            chain_strength (float/mapping/callable, optional):
                Sets the coupling strength between qubits representing variables 
                that form a :term:`chain`. Mappings should specify the required 
                chain strength for each variable. Callables should accept the BQM 
                and embedding and return a float or mapping. By default, 
                `chain_strength` is calculated with
                :func:`~dwave.embedding.chain_strength.uniform_torque_compensation`.

            chain_break_method (function, optional):
                Method used to resolve chain breaks during sample unembedding.
                See :func:`~dwave.embedding.unembed_sampleset`.

            chain_break_fraction (bool, optional, default=True):
                Add a ‘chain_break_fraction’ field to the unembedded response with
                the fraction of chains broken before unembedding.

            embedding_parameters (dict, optional):
                If provided, parameters are passed to the embedding method as
                keyword arguments. Overrides any `embedding_parameters` passed
                to the constructor. Only used on the first call.

            **parameters:
                Parameters for the sampling method, specified by the child
                sampler.

        Returns:
            :obj:`dimod.SampleSet`

        """
        if self.embedding is None:
            # get an embedding using the current find_embedding function
            embedding_parameters = parameters.pop('embedding_parameters', None)

            if embedding_parameters is None:
                embedding_parameters = self.embedding_parameters
            else:
                # update the base parameters with the new ones provided
                embedding_parameters.update((key, val)
                                            for key, val in self.embedding_parameters
                                            if key not in embedding_parameters)

            source_edgelist = list(itertools.chain(bqm.quadratic,
                                                   ((v, v) for v in bqm.linear)))

            target_edgelist = self.target_structure.edgelist

            embedding = self.find_embedding(source_edgelist, target_edgelist,
                                            **embedding_parameters)

            self._fix_embedding(embedding)

        return super(LazyFixedEmbeddingComposite, self).sample(bqm, **parameters)


class FixedEmbeddingComposite(LazyFixedEmbeddingComposite):
    """Maps problems to a structured sampler with the specified minor-embedding.

    Args:
        child_sampler (dimod.Sampler):
            Structured dimod sampler such as a D-Wave system.

        embedding (dict[hashable, iterable], optional):
            Mapping from a source graph to the specified sampler’s graph (the
            target graph).

        source_adjacency (dict[hashable, iterable]):
            Deprecated. Dictionary to describe source graph as `{node:
            {node neighbours}}`.

        kwargs:
            See the :class:`EmbeddingComposite` class for additional keyword
            arguments. Note that ``find_embedding`` and ``embedding_parameters``
            keyword arguments are ignored.

    Examples:
        To embed a triangular problem (a problem with a three-node complete graph,
        or clique) in the Chimera topology, you need to :term:`chain` two
        qubits. This example maps triangular problems to a composed sampler
        (based on the unstructured :class:`~dimod.reference.samplers.ExactSolver`)
        with a Chimera unit-cell structure.

        >>> import dimod
        >>> import dwave_networkx as dnx
        >>> from dwave.system import FixedEmbeddingComposite
        ...
        >>> c1 = dnx.chimera_graph(1)
        >>> embedding = {'a': [0, 4], 'b': [1], 'c': [5]}
        >>> structured_sampler = dimod.StructureComposite(dimod.ExactSolver(),
        ...                                               c1.nodes, c1.edges)
        >>> sampler = FixedEmbeddingComposite(structured_sampler, embedding)
        >>> sampler.edgelist
        [('a', 'b'), ('a', 'c'), ('b', 'c')]
    """
    def __init__(self, child_sampler, embedding=None, source_adjacency=None,
                 **kwargs):
        super(FixedEmbeddingComposite, self).__init__(child_sampler, **kwargs)

        # dev note: this entire block is to support a deprecated feature and can
        # be removed in the next major release
        if embedding is None:
            if source_adjacency is None:
                raise TypeError("either embedding or source_adjacency must be "
                                "provided")
            else:
                warn(("The source_adjacency parameter is deprecated"),
                     DeprecationWarning, stacklevel=2)

            source_edgelist = [(u, v) for u in source_adjacency for v in source_adjacency[u]]

            embedding = self.find_embedding(source_edgelist,
                                            self.target_structure.edgelist)

        self._fix_embedding(embedding)


class LazyEmbeddingComposite(LazyFixedEmbeddingComposite):
    """Deprecated. Maps problems to the structure of its first given problem.

    This class is deprecated; use the :class:`LazyFixedEmbeddingComposite` class instead.

    Args:
        sampler (dimod.Sampler):
            Structured dimod sampler.
    """
    def __init__(self, child_sampler):
        super(LazyEmbeddingComposite, self).__init__(child_sampler)
        warn("'LazyEmbeddingComposite' has been renamed to 'LazyFixedEmbeddingComposite'.", DeprecationWarning)


class AutoEmbeddingComposite(EmbeddingComposite):
    """Maps problems to a structured sampler, embedding if needed.

    This composite first tries to submit the binary quadratic model directly
    to the child sampler and only embeds if a
    :exc:`dimod.exceptions.BinaryQuadraticModelStructureError` is raised.

    Args:
        child_sampler (:class:`dimod.Sampler`):
            Structured dimod sampler, such as a
            :obj:`~dwave.system.samplers.DWaveSampler()`.

        find_embedding (function, optional):
            A function `find_embedding(S, T, **kwargs)` where `S` and `T`
            are edgelists. The function can accept additional keyword arguments.
            Defaults to :func:`minorminer.find_embedding`.

        kwargs:
            See the :class:`EmbeddingComposite` class for additional keyword
            arguments.

    """
    def __init__(self, child_sampler, **kwargs):

        child_search = kwargs.get('child_structure_search',
                                  dimod.child_structure_dfs)

        def permissive_child_structure(sampler):
            try:
                return child_search(sampler)
            except ValueError:
                return None
            except (AttributeError, TypeError):  # support legacy dimod
                return None

        super(AutoEmbeddingComposite, self).__init__(child_sampler,
                                                     child_structure_search=permissive_child_structure,
                                                     **kwargs)

    def sample(self, bqm, **parameters):
        child = self.child

        # we want to pass only the parameters relevent to the child sampler
        subparameters = {key: val for key, val in parameters.items()
                         if key in child.parameters}
        try:
            return child.sample(bqm, **subparameters)
        except dimod.exceptions.BinaryQuadraticModelStructureError:
            # does not match the structure so try embedding
            pass

        return super(AutoEmbeddingComposite, self).sample(bqm, **parameters)



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Utility functions for calculating chain strength.

Examples:
    This example uses :func:`uniform_torque_compensation`, given a prefactor of 2, 
    to calculate a chain strength that :class:`EmbeddingComposite` then uses.

    >>> from functools import partial
    >>> from dwave.system import EmbeddingComposite, DWaveSampler
    >>> from dwave.embedding.chain_strength import uniform_torque_compensation
    ...
    >>> Q = {(0,0): 1, (1,1): 1, (2,3): 2, (1,2): -2, (0,3): -2}
    >>> sampler = EmbeddingComposite(DWaveSampler())
    >>> # partial() can be used when the BQM or embedding is not accessible
    >>> chain_strength = partial(uniform_torque_compensation, prefactor=2)
    >>> sampleset = sampler.sample_qubo(Q, chain_strength=chain_strength, return_embedding=True)
    >>> sampleset.info['embedding_context']['chain_strength']
    1.224744871391589

"""
import math

__all__ = ['uniform_torque_compensation', 'scaled']

def uniform_torque_compensation(bqm, embedding=None, prefactor=1.414):
    """Chain strength that attempts to compensate for torque that would break
    the chain.

    The RMS of the problem's quadratic biases is used for calculation.

    Args: 
        bqm (:obj:`.BinaryQuadraticModel`):
            A binary quadratic model.

        embedding (dict/:class:`.EmbeddedStructure`, default=None):
            Included to satisfy the `chain_strength` callable specifications 
            for `embed_bqm`. 

        prefactor (float, optional, default=1.414):
            Prefactor used for scaling. For non-pathological problems, the recommended 
            range of prefactors to try is [0.5, 2].

    Returns:
        float: The chain strength, or 1 if chain strength is not applicable.

    """
    if bqm.num_interactions > 0:
        squared_j = (j ** 2 for j in bqm.quadratic.values())
        rms = math.sqrt(sum(squared_j)/bqm.num_interactions)
        avg_degree = bqm.degrees(array=True).mean()

        return prefactor * rms * math.sqrt(avg_degree)
    else:
        return 1    # won't matter (chain strength isn't needed to embed this problem)

def scaled(bqm, embedding=None, prefactor=1.0):
    """Chain strength that is scaled to the problem bias range.

    Args:
        bqm (:obj:`.BinaryQuadraticModel`):
            A binary quadratic model.

        embedding (dict/:class:`.EmbeddedStructure`, default=None):
            Included to satisfy the `chain_strength` callable specifications 
            for `embed_bqm`. 

        prefactor (float, optional, default=1.0):
            Prefactor used for scaling. 

    Returns:
        float: The chain strength, or 1 if chain strength is not applicable.

    """  
    if bqm.num_interactions > 0:
        max_bias = max(max(bqm.linear.max(), -bqm.linear.min()), 
                       max(bqm.quadratic.max(), -bqm.quadratic.min()))
        return prefactor * max_bias
    else:
        return 1    # won't matter (chain strength isn't needed to embed this problem)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""who watches the watchmen?"""
import os
import unittest
from unittest import mock

import numpy as np
import dimod
import dimod.testing as dit

from dwave.system import DWaveSampler
from dwave.system.testing import MockDWaveSampler, MockLeapHybridDQMSampler
from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError
from dimod import DiscreteQuadraticModel, ExtendedVartype, SampleSet


class TestMockDWaveSampler(unittest.TestCase):
    @unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
    def test_properties_and_params(self):
        try:
            sampler = DWaveSampler()
        except (ValueError, ConfigFileError, SolverNotFoundError):
            raise unittest.SkipTest("no QPU available")

        mock = MockDWaveSampler()

        self.assertEqual(set(mock.properties), set(sampler.properties))
        self.assertEqual(set(mock.parameters), set(sampler.parameters))

        #Check extraction of nodelist, edgelist and properties from 
        mock = MockDWaveSampler.from_qpu_sampler(sampler)
        self.assertEqual(mock.nodelist.sort(),sampler.nodelist.sort())
        self.assertEqual(mock.edgelist.sort(),sampler.edgelist.sort())

    def test_sampler(self):
        sampler = MockDWaveSampler()
        dit.assert_sampler_api(sampler)
        dit.assert_structured_api(sampler)

    def test_mock_parameters(self):
        # Single cell Chimera (DW2000Q) solver:
        sampler = MockDWaveSampler(topology_type='chimera', topology_shape = [1,1,4])
        num_reads = 10
        ss = sampler.sample_ising({0: -1}, {}, num_reads=num_reads)
        self.assertEqual(sum(ss.record.num_occurrences),num_reads)
        self.assertTrue(len(ss.record.num_occurrences) <= 2) #Binary state histogram
        ss = sampler.sample_ising({0: -1}, {}, num_reads=num_reads,
                                  answer_mode='raw')
        self.assertEqual(len(ss.record.num_occurrences), num_reads)
        
        ss = sampler.sample_ising({0: -1}, {}, num_reads=num_reads,
                                  answer_mode='raw', max_answers=2)
        self.assertEqual(len(ss.record.num_occurrences), 2)

        # note: greedy should be available; it's a package/test requirement
        # disable exact ground state calc
        with mock.patch.object(sampler, 'exact_solver_cutoff', 0):
            #QPU format initial states:
            initial_state = [(i,1) if i%4==0 else (i,3) for i in range(8)]
            ss = sampler.sample_ising({0: 1, 4: 1}, {(0, 4): -2},
                                      num_reads=1,
                                      answer_mode='raw',
                                      initial_state=initial_state)
            #The initialized state is a local minima, and should
            #be returned under greedy descent mocking:
            self.assertEqual(ss.record.energy[0],0)

    def test_unmock_parameters(self):
        # Check valid DWaveSampler() parameter, that is not mocked throws a warning:
        sampler = MockDWaveSampler()
        with self.assertWarns(UserWarning) as w:
            ss = sampler.sample_ising({0 : -1}, {}, annealing_time=123)
        with self.assertRaises(ValueError) as e:
            ss = sampler.sample_ising({0 : -1}, {}, monkey_time=123)
        with self.assertRaises(ValueError) as e:
            sampler = MockDWaveSampler(topology_type="chimera123")
        sampler = MockDWaveSampler(parameter_warnings=False)
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter(action="error", category=UserWarning)
            ss = sampler.sample_ising({0 : -1}, {}, annealing_time=123)

    def test_ground_override(self):
        # bqm with a local minimum at (-1, -1)
        bqm = dimod.BQM.from_ising({'a': -1, 'b': -1}, {'ab': -2})
        local_minimum = [-1, -1]    # energy: 0
        ground_state = [1, 1]       # energy: -4
        local_minimum_state = list(zip(bqm.variables, local_minimum))

        # simple two connected variables sampler (path(2))
        sampler = MockDWaveSampler(nodelist=['a', 'b'], edgelist=[('a', 'b')],
                                   exact_solver_cutoff=len(bqm.variables))

        # disable exact ground state calc and start from a local minimum -> greedy should stay stuck
        with mock.patch.object(sampler, 'exact_solver_cutoff', 0):
            ss = sampler.sample(bqm, initial_state=local_minimum_state)
            np.testing.assert_array_equal(ss.record[0].sample, local_minimum)

        # boundary on which exact ground state calc kicks in
        with mock.patch.object(sampler, 'exact_solver_cutoff', len(bqm.variables)):
            ss = sampler.sample(bqm, initial_state=local_minimum_state)
            np.testing.assert_array_equal(ss.record[0].sample, ground_state)

        # double-check the default
        self.assertEqual(sampler.exact_solver_cutoff, len(bqm.variables))
        ss = sampler.sample(bqm)
        np.testing.assert_array_equal(ss.record[0].sample, ground_state)

    def test_empty_bqm(self):
        sampler = MockDWaveSampler()
        bqm = dimod.BQM('SPIN')
        ss = sampler.sample(bqm)
        self.assertIs(ss.vartype, bqm.vartype)

    def test_chimera_topology(self):
        grid_parameter = 5
        tile_parameter = 2
        sampler = MockDWaveSampler(topology_type='chimera',
                                   topology_shape=[grid_parameter,
                                                   grid_parameter,
                                                   tile_parameter])
        # C5 (shore 2) fabric only has 200 nodes
        self.assertTrue(len(sampler.nodelist)==
                        grid_parameter*grid_parameter*tile_parameter*2)
        dit.assert_sampler_api(sampler)
        dit.assert_structured_api(sampler)

    def test_pegasus_topology(self):
        grid_parameter = 4
        sampler = MockDWaveSampler(topology_type='pegasus',
                                   topology_shape=[grid_parameter])
        # P4 fabric only has 264 nodes
        self.assertTrue(len(sampler.nodelist)==264)
        dit.assert_sampler_api(sampler)
        dit.assert_structured_api(sampler)

    def test_zephyr_topology(self):
        grid_parameter = 3
        tile_parameter = 4
        sampler = MockDWaveSampler(topology_type='zephyr',
                                   topology_shape=[grid_parameter,
                                                   tile_parameter])
        # P4 fabric only has 264 nodes
        self.assertTrue(len(sampler.nodelist)==
                        tile_parameter*grid_parameter*(8*grid_parameter + 4))
        dit.assert_sampler_api(sampler)
        dit.assert_structured_api(sampler)

    def test_properties(self):
        properties = {'topology' : {
            'type' : 'pegasus',
            'shape' : [5]}}
        # Note, topology_type and topology_shape must be consistent
        # or None.
        with self.assertRaises(ValueError) as e:
            sampler = MockDWaveSampler(topology_type='chimera',
                                       properties=properties)
        with self.assertRaises(ValueError) as e:
            sampler = MockDWaveSampler(topology_shape=[4],
                                       properties=properties)
        sampler = MockDWaveSampler(properties=properties)
        self.assertEqual(sampler.properties['topology']['type'],'pegasus')
        self.assertEqual(sampler.properties['topology']['shape'][-1],5)
        properties['category'] = 'test_choice'
        sampler = MockDWaveSampler(properties=properties)
        self.assertEqual(sampler.properties['category'],'test_choice')
        
    def test_yield_arguments(self):
        # Request 1 node and 1 edge deletion, check resulting graph 
        #    1      3
        #  X      2
        #
        #    5 -X-  7
        #  4      6
        # Defect free case: 8 nodes. 4 external edges, 4 internal edges
        # Delete first node (2 edges, 1 node)
        # Delete final external edge (1 edge)
    
        delete_nodes = [0]
        delete_edges = [(5, 7)]
        
        chimera_shape = [2, 2, 1]
        sampler = MockDWaveSampler(topology_type='chimera',
                                   topology_shape=chimera_shape,
                                   broken_nodes=delete_nodes,
                                   broken_edges=delete_edges)
        self.assertTrue(len(sampler.nodelist)==7)
        self.assertTrue(len(sampler.edgelist)==5)

        nodelist = [0,4,5,6,7]
        edgelist = [(5,7),(4,6)]
        sampler = MockDWaveSampler(topology_type='chimera',
                                   topology_shape=chimera_shape,
                                   nodelist=nodelist,
                                   edgelist=edgelist)
        self.assertTrue(len(sampler.nodelist)==5)
        self.assertTrue(len(sampler.edgelist)==2)
        
        sampler = MockDWaveSampler(topology_type='chimera',
                                   topology_shape=chimera_shape,
                                   nodelist=nodelist,
                                   edgelist=edgelist,
                                   broken_nodes=delete_nodes,
                                   broken_edges=delete_edges)
        self.assertTrue(len(sampler.nodelist)==4)
        self.assertTrue(len(sampler.edgelist)==1)
        
class TestMockLeapHybridDQMSampler(unittest.TestCase):
    def test_sampler(self):
        sampler = MockLeapHybridDQMSampler()

        self.assertTrue(callable(sampler.sample_dqm))
        self.assertTrue(callable(sampler.min_time_limit))
        self.assertTrue(hasattr(sampler, 'properties'))
        self.assertTrue(hasattr(sampler, 'parameters'))

        dqm = DiscreteQuadraticModel()
        dqm.add_variable(3)
        dqm.add_variable(4)

        result = sampler.sample_dqm(dqm)
        self.assertTrue(isinstance(result, SampleSet))
        self.assertGreaterEqual(len(result), 12)  # min num of samples from dqm solver
        self.assertEqual(len(result.variables), 2)
        self.assertEqual(result.vartype, ExtendedVartype.DISCRETE)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
from uuid import uuid4
from unittest import mock
from concurrent.futures import Future

import numpy as np
from networkx.utils import graphs_equal
from parameterized import parameterized

import dimod
import dwave_networkx as dnx

from dwave.cloud import exceptions
from dwave.cloud import computation

from dwave.system.samplers import DWaveSampler, qpu_graph
from dwave.system.warnings import EnergyScaleWarning, TooFewSamplesWarning
from dwave.system.exceptions import FailoverCondition, RetryCondition


C16 = dnx.chimera_graph(16)

# remove one node from C16 to simulate a not-fully-yielded system
C16.remove_node(42)

edges = set(tuple(edge) for edge in C16.edges)
edges.update([(v, u) for u, v in edges])  # solver has bi-directional


class MockSolver():
    nodes = set(range(2048))
    edges = edges
    properties = {'readout_thermalization_range': [0.0, 10000.0],
                  'annealing_time_range': [1.0, 2000.0],
                  'default_readout_thermalization': 0.0,
                  'parameters': {'num_spin_reversal_transforms': '',
                                 'programming_thermalization': '',
                                 'anneal_offsets': '',
                                 'num_reads': '',
                                 'max_answers': '',
                                 'readout_thermalization': '',
                                 'beta': '',
                                 'answer_mode': '',
                                 'auto_scale': '',
                                 'postprocess': '',
                                 'anneal_schedule': '',
                                 'chains': ''},
                  'chip_id': 'MockSolver'}

    def sample_bqm(self, bqm, num_reads=1, **kwargs):
        problem_id = str(uuid4())
        problem_label = kwargs.pop('label', None)
        info = dict(timing={'total_real_time': 11511.0, 'anneal_time_per_run': 20.0,
                            'post_processing_overhead_time': 2042.0,
                            'qpu_sampling_time': 164.0,
                            'readout_time_per_run': 123.0,
                            'qpu_delay_time_per_sample': 21.0,
                            'qpu_anneal_time_per_sample': 20.0,
                            'total_post_processing_time': 2042.0,
                            'qpu_programming_time': 8740.0,
                            'run_time_chip': 164.0,
                            'qpu_access_time': 11511.0,
                            'qpu_readout_time_per_sample': 123.0},
                    problem_id=problem_id)
        if problem_label:
            info.update(problem_label=problem_label)

        samples = np.random.choice(tuple(bqm.vartype.value),
                                   size=(num_reads, len(bqm)))

        future = Future()
        ss = dimod.SampleSet.from_samples_bqm((samples, bqm.variables), bqm,
                                              info=info)
        future.sampleset = ss
        future.id = problem_id
        future.label = problem_label

        return future


class TestDWaveSampler(unittest.TestCase):
    @mock.patch('dwave.system.samplers.dwave_sampler.Client')
    def setUp(self, MockClient):

        # using the mock
        self.sampler = DWaveSampler()

        self.sampler.solver = MockSolver()

    @mock.patch('dwave.system.samplers.dwave_sampler.Client')
    def test_init_default(self, MockClient):
        """QPU with the highest number of qubits chosen by default."""

        sampler = DWaveSampler()

        MockClient.from_config.assert_called_once_with(
            client='qpu',
            defaults={'solver': {'order_by': '-num_active_qubits'}})

    @mock.patch('dwave.system.samplers.dwave_sampler.Client')
    def test_init_generic_behavior(self, MockClient):
        """Generic solver behavior (default prior to 0.10.0) can be forced."""

        sampler = DWaveSampler(client='base')

        MockClient.from_config.assert_called_once_with(
            client='base',
            defaults={'solver': {'order_by': '-num_active_qubits'}})

    @mock.patch('dwave.system.samplers.dwave_sampler.Client')
    def test_init_solver(self, MockClient):
        """QPU can be explicitly selected (old default usage example)"""

        solver = {'qpu': True, 'num_qubits__gt': 1000}

        sampler = DWaveSampler(solver=solver)

        MockClient.from_config.assert_called_once_with(
            client='qpu',
            solver=solver,
            defaults={'solver': {'order_by': '-num_active_qubits'}})

    def test_sample_ising_variables(self):

        sampler = self.sampler

        response = sampler.sample_ising({0: -1, 1: 1}, {})

        rows, cols = response.record.sample.shape

        self.assertEqual(cols, 2)

        response = sampler.sample_ising({}, {(0, 4): 1})

        rows, cols = response.record.sample.shape

        self.assertEqual(cols, 2)

        self.assertFalse(np.any(response.record.sample == 0))
        self.assertIs(response.vartype, dimod.SPIN)

        self.assertIn('num_occurrences', response.record.dtype.fields)
        self.assertIn('timing', response.info)
        self.assertIn('problem_id', response.info)

    def test_sample_qubo_variables(self):

        sampler = self.sampler

        response = sampler.sample_qubo({(0, 0): -1, (1, 1): 1})

        rows, cols = response.record.sample.shape

        self.assertEqual(cols, 2)

        response = sampler.sample_qubo({(0, 0): -1, (1, 1): 1})

        rows, cols = response.record.sample.shape

        self.assertEqual(cols, 2)

        self.assertTrue(np.all(response.record.sample >= 0))
        self.assertIs(response.vartype, dimod.BINARY)

        self.assertIn('num_occurrences', response.record.dtype.fields)
        self.assertIn('timing', response.info)
        self.assertIn('problem_id', response.info)

    def test_problem_labelling(self):
        sampler = self.sampler

        # label parameter is supported
        self.assertIn('label', sampler.parameters)

        # no-label case works as before
        ss = sampler.sample_ising({}, {(0, 4): 1})

        self.assertIn('problem_id', ss.info)
        self.assertNotIn('problem_label', ss.info)

        # label is propagated to sampleset.info
        label = 'problem label'
        ss = sampler.sample_ising({}, {(0, 4): 1}, label=label)

        self.assertIn('problem_id', ss.info)
        self.assertIn('problem_label', ss.info)
        self.assertEqual(ss.info.get('problem_label'), label)

    @mock.patch('dwave.system.samplers.dwave_sampler.Client')
    def test_failover_off(self, MockClient):
        sampler = DWaveSampler(failover=False)

        sampler.solver.sample_bqm.side_effect = exceptions.SolverOfflineError

        with self.assertRaises(exceptions.SolverOfflineError):
            sampler.sample_ising({}, {})

    @parameterized.expand([
        (exceptions.InvalidAPIResponseError, FailoverCondition),
        (exceptions.SolverNotFoundError, FailoverCondition),
        (exceptions.SolverOfflineError, FailoverCondition),
        (exceptions.SolverError, FailoverCondition),
        (exceptions.PollingTimeout, RetryCondition),
        (exceptions.SolverAuthenticationError, exceptions.SolverAuthenticationError),   # auth error propagated
        (KeyError, KeyError),   # unrelated errors propagated
    ])
    @mock.patch('dwave.system.samplers.dwave_sampler.Client')
    def test_async_failover(self, source_exc, target_exc, MockClient):
        sampler = DWaveSampler(failover=True)

        mocksolver = sampler.solver
        edgelist = sampler.edgelist

        # call once (async, no need to resolve)
        sampler.sample_ising({}, {})

        self.assertIs(mocksolver, sampler.solver)  # still same solver

        # one of the sample methods was called
        self.assertEqual(sampler.solver.sample_ising.call_count
                         + sampler.solver.sample_qubo.call_count
                         + sampler.solver.sample_bqm.call_count, 1)

        # simulate solver exception on sampleset resolve
        fut = computation.Future(mocksolver, None)
        fut._set_exception(source_exc)
        sampler.solver.sample_bqm = mock.Mock()
        sampler.solver.sample_bqm.return_value = fut

        # verify failover signalled
        with self.assertRaises(target_exc):
            sampler.sample_ising({}, {}).resolve()

        # make sure get_solver makes a new mock solver
        sampler.client.get_solver.reset_mock(return_value=True)

        # trigger failover
        sampler.trigger_failover()

        # verify failover
        self.assertIsNot(mocksolver, sampler.solver)  # new solver
        self.assertIsNot(edgelist, sampler.edgelist)  # also should be new

    def test_warnings_energy_range(self):
        sampler = self.sampler

        response = sampler.sample_ising({0: 10**-4, 1: 1}, {},
                                        warnings='SAVE')

        self.assertIn('warnings', response.info)
        count = 0
        for warning in response.info['warnings']:
            if issubclass(warning['type'], EnergyScaleWarning):
                count += 1
        self.assertEqual(count, 1)

        response = sampler.sample_qubo({(0, 4): -1}, warnings='SAVE')

        count = 0
        for warning in response.info['warnings']:
            if issubclass(warning['type'], EnergyScaleWarning):
                count += 1
        self.assertEqual(count, 1)

    def test_warnings_sampling_error(self):
        sampler = self.sampler

        response = sampler.sample_ising({0: 10**-4, 1: 1}, {},
                                        warnings='SAVE')

        self.assertIn('warnings', response.info)
        count = 0
        for warning in response.info['warnings']:
            if issubclass(warning['type'], TooFewSamplesWarning):
                count += 1
        self.assertEqual(count, 1)
        self.assertEqual(len(response.info['warnings']), 2)

        response = sampler.sample_qubo({(0, 4): -1}, warnings='SAVE')

        count = 0
        for warning in response.info['warnings']:
            if issubclass(warning['type'], TooFewSamplesWarning):
                count += 1
        self.assertEqual(count, 1)
        self.assertEqual(len(response.info['warnings']), 2)

    def test_to_networkx_chimera(self):
        sampler = self.sampler
        sampler.solver.properties.update({'topology': {'type': 'chimera', 'shape': [4, 4, 4]}})
        G = sampler.to_networkx_graph()
        G2 = qpu_graph(sampler.properties['topology']['type'],
                       sampler.properties['topology']['shape'],
                       sampler.nodelist,
                       sampler.edgelist)
        self.assertTrue(graphs_equal(G,G2))
        # Create chimera graph for comparison
        chimeraG = dnx.chimera_graph(4, node_list=sampler.nodelist, edge_list=sampler.edgelist)

        self.assertEqual(set(G), set(chimeraG))

        for u, v in chimeraG.edges:
            self.assertIn(u, G[v])
            
        del sampler.solver.properties['topology']

    def test_to_networkx_pegasus(self):
        sampler = self.sampler
        sampler.solver.properties.update({'topology': {'type': 'pegasus', 'shape': [4]}})
        G = sampler.to_networkx_graph()

        # Create pegasus graph for comparison
        pegasusG = dnx.pegasus_graph(4, node_list=sampler.nodelist, edge_list=sampler.edgelist)

        self.assertEqual(set(G), set(pegasusG))

        for u, v in pegasusG.edges:
            self.assertIn(u, G[v])

        del sampler.solver.properties['topology']

    def test_to_networkx_zephyr(self):
        sampler = self.sampler
        sampler.solver.properties.update({'topology': {'type': 'zephyr', 'shape': [4, 4]}})
        G = sampler.to_networkx_graph()

        # Create zephyr graph for comparison
        zephyrG = dnx.zephyr_graph(4, node_list=sampler.nodelist, edge_list=sampler.edgelist)

        self.assertEqual(set(G), set(zephyrG))

        for u, v in zephyrG.edges:
            self.assertIn(u, G[v])

        del sampler.solver.properties['topology']

class TestDWaveSamplerAnnealSchedule(unittest.TestCase):
    def test_typical(self):
        class MockScheduleSampler(DWaveSampler):
            parameters = {'anneal_schedule': ''}
            properties = {'max_anneal_schedule_points': 4,
                          'annealing_time_range': [1.0, 2000.0]}

            def __init__(self):
                pass

        DWaveSampler.validate_anneal_schedule(MockScheduleSampler(), [[0.0, 0.0], [0.2, 0.2], [5.2, 0.2], [6.0, 1.0]])
        DWaveSampler.validate_anneal_schedule(MockScheduleSampler(), [(0, 1), (55.0, 0.45), (155.0, 0.45), (210.0, 1)])



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import dwave_networkx as dnx

from dwave.cloud.testing import isolated_environ

from dwave.system import common_working_graph
from dwave.system.utilities import FeatureFlags


class TestCommonWorkingGraph(unittest.TestCase):
    def test_single_tile(self):

        G1 = dnx.chimera_graph(1)
        G = common_working_graph(G1, G1)

        # should have 8 nodes
        self.assertEqual(len(G), 8)

        # nodes 0,...,7 should be in the tile
        for n in range(8):
            self.assertIn(n, G)

        # check bipartite
        for i in range(4):
            for j in range(4, 8):
                self.assertTrue((i, j) in G.edges() or (j, i) in G.edges())

    def test_c1_c2_tiles(self):
        G1 = dnx.chimera_graph(1)
        G2 = dnx.chimera_graph(2)

        G = common_working_graph(G1, G1)

        self.assertEqual(len(G), 8)

    def test_missing_node(self):
        G1 = dnx.chimera_graph(1)
        G1.remove_node(2)
        G2 = dnx.chimera_graph(2)

        G = common_working_graph(G1, G1)

        self.assertNotIn(2, G)
        self.assertNotIn((2, 4), G.edges())

    def test_sampler_adjacency(self):
        adj = {0: {1, 2}, 1: {2}, 2: {0, 1}}
        G = dnx.chimera_graph(1)

        H = common_working_graph(adj, G)

        self.assertEqual(set(H.nodes), {0, 1, 2})
        self.assertEqual(set(H.edges), set())


class TestFeatureFlagSupport(unittest.TestCase):
    def test_base(self):
        with isolated_environ(add={'DWAVE_FEATURE_FLAGS': '{"x": true}'}):
            self.assertTrue(FeatureFlags.get('x'))
        with isolated_environ(add={'DWAVE_FEATURE_FLAGS': '{"x": false}'}):
            self.assertFalse(FeatureFlags.get('x'))
        with isolated_environ(add={'DWAVE_FEATURE_FLAGS': '{"x": 0}'}):
            self.assertFalse(FeatureFlags.get('x'))
        with isolated_environ(add={'DWAVE_FEATURE_FLAGS': '{"x": 1}'}):
            self.assertTrue(FeatureFlags.get('x'))
        with isolated_environ(add={'DWAVE_FEATURE_FLAGS': '{"x": error}'}):
            self.assertFalse(FeatureFlags.get('x'))
        with isolated_environ(add={'DWAVE_FEATURE_FLAGS': '{"y": true}'}):
            self.assertFalse(FeatureFlags.get('x'))
        with isolated_environ(add={'DWAVE_FEATURE_FLAGS': ''}):
            self.assertFalse(FeatureFlags.get('x'))
        with isolated_environ(remove=('DWAVE_FEATURE_FLAGS',)):
            self.assertFalse(FeatureFlags.get('x'))

    def test_hss_solver_config_override(self):
        with isolated_environ(add={'DWAVE_FEATURE_FLAGS': '{"hss_solver_config_override": true}'}):
            self.assertTrue(FeatureFlags.hss_solver_config_override)
        with isolated_environ(remove=('DWAVE_FEATURE_FLAGS',)):
            self.assertFalse(FeatureFlags.hss_solver_config_override)
